/*
<H2>Licence Agreement</H2>
<PRE>InterfaceClassFinder package
Copyright (c) 2009 Forward Computing and Control Pty. Ltd.
NSW. Australia,  www.forward.com.au
All rights reserved.

Redistribution of the source of this package, with or without
modification, is NOT permitted.

Redistribution of the compiled java .class files of this package, with or without
modification, are permitted provided that the following conditions
are met:
1. Redistributions in compiled java .class form must reproduce the above
copyright notice, this list of conditions and the following disclaimer
in the documentation and/or other materials provided with the distribution.

2. If the software is modified, the compiled java .class distribution must
include, in the documentation and/or other materials provided with the
distribution, a statement that the software has been modified and
identify the person who was responsable for the modifications.

3. Neither the name of the author nor Forward Computing and Control may be
used to endorse or promote products derived from this software
without specific prior written permission.

4.  All advertising materials mentioning features or use of this
software must display the following acknowledgment:
"This product includes software developed by Matthew Ford
and Forward Computing and Control Pty. Ltd.
(http://www.forward.com.au)."


THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
</PRE>
 */
package au.com.forward.interfaceClassFinder;

import java.io.File;
import java.io.FileInputStream;
import java.io.InputStream;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.Iterator;
import java.util.List;
import java.util.jar.JarEntry;
import java.util.jar.JarInputStream;
import org.objectweb.asm.ClassReader;

/**
 * This class finds all classes that implement a given interface and returns the class names as a list.<br>
 * Java libraries are not scanned. Only the classes and jars on the class path are scanned.
 * @author Matthew
 */
public class FindInterfaceClasses {

    public static String newline = System.getProperty("line.separator");
    private static String interfaceClassName = "";
    private static HashMap<String, ClassInfo> map = new HashMap<String, ClassInfo>();
    private static boolean mapImplementsNulled = true; // initially all the implementsInterface settings are null
    private static final IClasspathProcessor consumer = new IClasspathProcessor() {

        private ClassVisitor classVisitor = new ClassVisitor();

        public void processClass(String className, String rootDir, File classFile) {
            try {
                InputStream is = new FileInputStream(classFile);
                ClassReader cr = new ClassReader(is);
                cr.accept(classVisitor, 0);
                ClassInfo classInfo = classVisitor.getClassInfo();
                ClassInfo existingClassInfo = map.get(classInfo.getClassName());
                if (existingClassInfo != null) {
                    //"Duplicate class files " + newline + existingClassInfo.asString() + newline + classInfo.asString());
                    // always keep the first one found
                } else {
                    // not there add it
                    map.put(classInfo.getClassName(), classInfo);
                }
            } catch (Throwable t) {
                throw new RuntimeException("Error checking " + classFile.getPath() + " " + t.toString(), t);
            }
        }

        public void processClass(String className, JarEntry jarEntry, File jarFile) {
            try {
                JarInputStream jarInputStream = new JarInputStream(new FileInputStream(jarFile), false);
                JarEntry entry = jarInputStream.getNextJarEntry();
                while (entry != null) {
                    if (entry.getName().equals(jarEntry.getName())) {
                        break; // found it
                    }
                    entry = jarInputStream.getNextJarEntry();
                }
                if (entry == null) {
                    throw new IllegalArgumentException("Jar file:" + jarFile.getPath() + " does contain entry:" + jarEntry.getName());
                }
                ClassReader cr = new ClassReader(jarInputStream);
                cr.accept(classVisitor, 0);
                ClassInfo classInfo = classVisitor.getClassInfo();
                // set source for jar files
                classInfo.setSource(jarFile.getPath());
                ClassInfo existingClassInfo = map.get(classInfo.getClassName());
                if (existingClassInfo != null) {
                    //"Duplicate class files " + newline + existingClassInfo.asString() + newline + classInfo.asString());
                    // always keep the first one found
                } else {
                    // not there add it
                    map.put(classInfo.getClassName(), classInfo);
                }
            } catch (Throwable t) {
                throw new RuntimeException("Error checking " + jarEntry.getName() + " in " + jarFile.getPath() + " " + t.toString(), t);
            }
        }
    };

    /**
     * Scans the classpath and returns a list of fully qualified class names that implement the given interface class.<br>
     * Only classes which implement the interface and have a no-arg public constructor are returned.<br>
     * The interface can be implemented by a super class or the class can implement an interface that extends the given one.
     *
     * @param interfaceClass  the interface class to look for
     * @return a list of fully qualified class names that implement the given interface and have a no-arg public constructor
     */
    public static List<String> findClassNames(Class interfaceClass) {
        ClasspathScanner<IClasspathProcessor> scanner = new ClasspathScanner<IClasspathProcessor>();
        map.clear();
        scanner.addConsumer(consumer);
        scanner.scanClassPath(); // builds up map
        List<String> classes = getClassNames(interfaceClass);
        return classes; // found these
    }

    /**
     * Clears the implementsInterface flag for all classes and interfaces in the classpath map.
     */
    private static void clearMapImplementsInterface() {
        if (mapImplementsNulled) {
            return; // already done
        }
        Iterator<ClassInfo> it = map.values().iterator();
        while (it.hasNext()) {
            ClassInfo classToCheck = it.next();
            classToCheck.clearImplementsInterface();
        }
        mapImplementsNulled = true; // all nulled now
    }

    /**
     * Having called findClassNames() once you can then call this method with another interface class
     * to find the classes that implement that interface.
     * 
     * @param interfaceClass the interface the classes must implement.
     * @return a list of fully qualified class names that implement the given interface and have a no-arg public constructor
     */
    public static List<String> getClassNames(Class interfaceClass) {
        if (interfaceClass == null) {
            throw new IllegalArgumentException("interfaceClass cannot be null.");
        }
        if (map.isEmpty()) {
            throw new IllegalStateException("Need to call findClassNames() first to scan classpath.");
        }
        interfaceClassName = interfaceClass.getCanonicalName();
        if (interfaceClassName == null) {
            throw new IllegalArgumentException("interfaceClass cannot be local or anonymous class.");
        }
        // replace . with /
        interfaceClassName = interfaceClassName.replace('.', '/');
        clearMapImplementsInterface();
        mapImplementsNulled = false; // about to set these values
        // scan classes for ones to load
        List<String> classes = new ArrayList<String>();
        Iterator<ClassInfo> it = map.values().iterator();
        while (it.hasNext()) {
            ClassInfo classToCheck = it.next();
            if (implementsInterface(interfaceClassName, classToCheck)) {
                if (classToCheck.hasNoArgConstructor() && classToCheck.isPublicClass() && !classToCheck.isAbstractClass()) {
                    classes.add(classToCheck.getClassName().replace('/', '.'));
                }
            }
        // check interfaces for this one.
        }
        return classes;
    }

    /**
     * Recursively checks if this class or interface implements the required interface.<br>
     * Updates the map implementsInterface variable as it goes to short cut searching heirarchies more than once.
     * @param interfaceClassName the interface class name to look for.
     * @param classInfo the interface and super class information for the class being checked.
     * @param level the level of recursion, for pretty logging.
     * @return true if this class info being check, or one of its super-classes implements the given interface.
     */
    private static boolean implementsInterface(String interfaceClassName, ClassInfo classInfo) {
        if (classInfo.getClassName().equals(interfaceClassName)) {
            // this is the interface we are looking for
            // check that it is actually an interface
            if (!classInfo.isInterface()) {
                throw new RuntimeException("" + interfaceClassName + " is not an interface.");
            }
            classInfo.setImplementsInterface(true);
            return true;
        }
        for (String s : classInfo.getInterfaceNames()) {
            if (s.equals(interfaceClassName)) {
                classInfo.setImplementsInterface(true);
                return true;
            }
        }
        // else look at super class
        String superClassName = classInfo.getSuperClassName();
        ClassInfo superClassInfo = map.get(superClassName);
        if (superClassInfo != null) {
            // skip language classes and classes not found
            if (superClassInfo.implementsInterface() == null) {
                // not checked yet
                if (implementsInterface(interfaceClassName, superClassInfo)) {
                    return true;
                } // else need to check interfaces below
            } else {
                // already checked this branch
                if (superClassInfo.implementsInterface()) {
                    return true;
                } // else need to check interfaces below
            }
        }
        // else either no super class scanned or returned false
        // look at interfaces
        for (String s : classInfo.getInterfaceNames()) {
            ClassInfo interfaceClassInfo = map.get(s);
            if (interfaceClassInfo != null) {
                // skip language classes and classes not found
                if (interfaceClassInfo.implementsInterface() == null) {
                    if (implementsInterface(interfaceClassName, interfaceClassInfo)) {
                        return true;
                    }
                } else {
                    // already checked this branch
                    if (interfaceClassInfo.implementsInterface()) {
                        return true;
                    }
                }
            }
        }
        // else does not implement the interface either itself or its super-classes or its interfaces and their super-interfaces
        classInfo.setImplementsInterface(false);
        return false;
    }

    /**
     * Find the classes that implement the give fully qualified class and which have no-arg public constructors.
     * 
     * @param args the full qualified interface class name, must exist in the class.
     */
    public static void main(String[] args) {
        if (args.length != 1) {
            System.out.println("Usage: FindInterfaceClasses fullyQualifiedInterfaceName.");
            System.exit(1);
        }
        try {
            Class interfaceClass = Class.forName(args[0]);
            List<String> classes = findClassNames(interfaceClass);
            if (classes.isEmpty()) {
                System.out.println(" No classes with no-arg public constuctors were found that implemented " + interfaceClass.getCanonicalName());
            } else {
                System.out.println(" Found the following classes with no-arg public constructors that implement " + interfaceClass.getCanonicalName());
                for (String s : classes) {
                    System.out.println(s);
                }
                System.out.println(" ---------- ");
            }
        } catch (Throwable t) {
            t.printStackTrace();
        }
    }
}